"""Statistics CLI method."""
from pathlib import Path
from typing import Optional, Union

import click

from eddington.cli.common_flags import data_file_option, output_dir_option, sheet_option
from eddington.cli.main_cli import eddington_cli
from eddington.cli.util import load_data_file

# pylint: disable=too-many-arguments


@eddington_cli.command("statistics")
@click.pass_context
@data_file_option
@sheet_option
@output_dir_option
@click.option("-n", "--file-name", type=str, help="Output file name.")
@click.option(
    "-f",
    "--file-format",
    type=click.Choice(["csv", "xlsx"], case_sensitive=False),
    default="xlsx",
    help="Output file name.",
)
def eddington_statistics(
    ctx: click.Context,
    data_file: str,
    sheet: Optional[str],
    output_dir: Optional[Union[Path, str]],
    file_name: Optional[str],
    file_format: str,
):
    """Print statistics of given data file."""
    data = load_data_file(
        ctx,
        Path(data_file),
        sheet,
    )
    if output_dir is None:
        click.echo(f'Data statistics of "{data_file}"')
        for column in data.all_columns:
            click.echo(f"{column}:")
            column_statistics = data.statistics(column)
            click.echo(f"\tMean: {column_statistics.mean}")
            click.echo(f"\tMedian: {column_statistics.median}")
            click.echo(f"\tVariance: {column_statistics.variance}")
            click.echo(f"\tStandard Deviation: {column_statistics.standard_deviation}")
            click.echo(f"\tMinimum: {column_statistics.minimum_value}")
            click.echo(f"\tMaximum: {column_statistics.maximum_value}")
        return
    output_dir = Path(output_dir)
    if not output_dir.exists():
        output_dir.mkdir()
    if file_format == "csv":
        data.save_statistics_csv(output_directory=output_dir, name=file_name)
    if file_format == "xlsx":
        data.save_statistics_excel(output_directory=output_dir, name=file_name)



"""CLI for Eddington."""

from eddington.cli.fit import eddington_fit
from eddington.cli.fit_random import eddington_fit_random
from eddington.cli.list import eddington_list
from eddington.cli.main_cli import eddington_cli
from eddington.cli.statistics_cli import eddington_statistics

__all__ = [
    "eddington_cli",
    "eddington_statistics",
    "eddington_list",
    "eddington_fit",
    "eddington_fit_random",
]



"""Fit CLI method."""
from pathlib import Path
from typing import Optional, Union

import click

from eddington.cli.common_flags import (
    a0_option,
    data_file_option,
    fitting_function_argument,
    is_grid_option,
    is_json_option,
    is_legend_option,
    is_x_log_scale_option,
    is_y_log_scale_option,
    output_dir_option,
    polynomial_option,
    sheet_option,
    should_plot_data_option,
    should_plot_fitting_option,
    should_plot_residuls_option,
    title_option,
    x_label_option,
    y_label_option,
)
from eddington.cli.main_cli import eddington_cli
from eddington.cli.util import (
    extract_array_from_string,
    fit_and_plot,
    load_data_file,
    load_fitting_function,
)

# pylint: disable=invalid-name,too-many-arguments,too-many-locals,duplicate-code


@eddington_cli.command("fit")
@click.pass_context
@fitting_function_argument
@polynomial_option
@a0_option
@data_file_option
@sheet_option
@click.option("--x-column", type=str, help="Column to read x values from.")
@click.option("--xerr-column", type=str, help="Column to read x error values from.")
@click.option("--y-column", type=str, help="Column to read y values from.")
@click.option("--yerr-column", type=str, help="Column to read y error values from.")
@title_option
@x_label_option
@y_label_option
@is_grid_option
@should_plot_fitting_option
@should_plot_residuls_option
@should_plot_data_option
@is_legend_option
@is_x_log_scale_option
@is_y_log_scale_option
@output_dir_option
@is_json_option
def eddington_fit(
    ctx: click.Context,
    fitting_function_name: Optional[str],
    polynomial_degree: Optional[int],
    a0: Optional[str],
    data_file: str,
    sheet: Optional[str],
    x_column: Optional[str],
    xerr_column: Optional[str],
    y_column: Optional[str],
    yerr_column: Optional[str],
    title: Optional[str],
    x_label: Optional[str],
    y_label: Optional[str],
    grid,
    should_plot_fitting: bool,
    should_plot_residuals: bool,
    should_plot_data: bool,
    legend: Optional[bool],
    x_log_scale: bool,
    y_log_scale: bool,
    output_dir: Union[Path, str],
    json: bool,
):
    """Fitting data file according to a fitting function."""
    data = load_data_file(
        ctx,
        Path(data_file),
        sheet,
        x_column=x_column,
        xerr_column=xerr_column,
        y_column=y_column,
        yerr_column=yerr_column,
    )
    func = load_fitting_function(
        ctx=ctx, func_name=fitting_function_name, polynomial_degree=polynomial_degree
    )
    fit_and_plot(
        data=data,
        func=func,
        a0=extract_array_from_string(a0),
        legend=legend,
        output_dir=output_dir,
        is_json=json,
        title=title,
        x_label=x_label,
        y_label=y_label,
        should_plot_data=should_plot_data,
        should_plot_fitting=should_plot_fitting,
        should_plot_residuals=should_plot_residuals,
        x_log_scale=x_log_scale,
        y_log_scale=y_log_scale,
        grid=grid,
    )



"""Module for handling statistical values."""
from dataclasses import dataclass, fields
from typing import List, Union

import numpy as np


@dataclass
class Statistics:
    """Common statistics fields for data analysis."""

    mean: float
    median: float
    variance: float
    standard_deviation: float
    maximum_value: float
    minimum_value: float

    @classmethod
    def from_array(cls, values_array: Union[List[float], np.ndarray]):
        """
        Build statistics object for a given values array.

        :param values_array: Values to calculate statistics for.
        :type values_array: numpy.ndarray or lis
        :return: Statistics of the given array
        :rtype: Statistics
        """
        if not isinstance(values_array, np.ndarray):
            values_array = np.array(values_array)
        if values_array.shape[0] == 0:
            raise ValueError("Cannot calculate statistics of no values.")
        return Statistics(
            mean=np.average(values_array),
            median=np.median(values_array),
            variance=np.var(values_array),
            standard_deviation=np.std(values_array),
            maximum_value=np.max(values_array),
            minimum_value=np.min(values_array),
        )

    @classmethod
    def parameters(cls):
        """Get list of available statistics parameters."""
        return [field.name for field in fields(cls)]



"""List of common flags in use of the Eddington CLI."""
import click

# Fitting


data_file_option = click.option(
    "-d",
    "--data-file",
    required=True,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    help="Data file to read from.",
)
sheet_option = click.option(
    "-s", "--sheet", type=str, help="Sheet name for excel files."
)
fitting_function_argument = click.argument(
    "fitting_function_name", type=str, default=""
)
polynomial_option = click.option(
    "-p",
    "--polynomial",
    "polynomial_degree",
    type=int,
    help="Fitting data to polynomial of nth degree.",
)
a0_option = click.option(
    "--a0",
    type=str,
    help=(
        "Initial guess for the fitting algorithm. "
        "Should be given as floating point numbers separated by commas"
    ),
)


# Plot

title_option = click.option("--title", type=str, help="Title for the fitting plot.")

x_label_option = click.option("--x-label", type=str, help="Label for the x axis.")

y_label_option = click.option("--y-label", type=str, help="Label for the y axis.")

is_grid_option = click.option(
    "--grid/--no-grid", default=False, help="Add grid lines to plots."
)
should_plot_fitting_option = click.option(
    "--plot-fitting/--no-plot-fitting",
    "should_plot_fitting",
    default=True,
    help="Should plot fitting.",
)

should_plot_residuls_option = click.option(
    "--plot-residuals/--no-plot-residuals",
    "should_plot_residuals",
    default=True,
    help="Should plot residuals.",
)

should_plot_data_option = click.option(
    "--plot-data/--no-plot-data",
    "should_plot_data",
    default=False,
    help="Should plot data.",
)

is_legend_option = click.option(
    "--legend/--no-legend",
    default=None,
    help="Should add legend to fitting plot.",
)

is_x_log_scale_option = click.option(
    "--x-log-scale/--no-x-log-scale",
    default=False,
    help="Change x axis scale to logarithmic.",
)

is_y_log_scale_option = click.option(
    "--y-log-scale/--no-y-log-scale",
    default=False,
    help="Change y axis scale to logarithmic.",
)


# Output
output_dir_option = click.option(
    "-o",
    "--output-dir",
    type=click.Path(dir_okay=True, file_okay=False),
    help="Output directory to save plots in.",
)

is_json_option = click.option(
    "--json",
    is_flag=True,
    default=False,
    help="Save result as json instead of text.",
)



"""Fitting data class insert the fitting algorithm."""
import collections
import csv
import json
from collections import OrderedDict, namedtuple
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import openpyxl

from eddington import io_util
from eddington.consts import (
    DEFAULT_MAX_COEFF,
    DEFAULT_MEASUREMENTS,
    DEFAULT_MIN_COEFF,
    DEFAULT_XMAX,
    DEFAULT_XMIN,
    DEFAULT_XSIGMA,
    DEFAULT_YSIGMA,
)
from eddington.exceptions import (
    FittingDataColumnExistenceError,
    FittingDataColumnIndexError,
    FittingDataColumnsLengthError,
    FittingDataColumnsSelectionError,
    FittingDataInvalidFile,
    FittingDataSetError,
)
from eddington.random_util import random_array, random_error, random_sigma
from eddington.statistics import Statistics

Columns = namedtuple("ColumnsResult", ["x", "y", "xerr", "yerr"])


class FittingData:  # pylint: disable=R0902,R0904
    """Fitting data class."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        data: Union[OrderedDict, Dict[str, np.ndarray]],
        x_column: Optional[Union[str, int]] = None,
        xerr_column: Optional[Union[str, int]] = None,
        y_column: Optional[Union[str, int]] = None,
        yerr_column: Optional[Union[str, int]] = None,
    ):
        """
        Constructor.

        :param data: Dictionary from a column name to its values
        :type data: ``dict`` or ``OrderedDict`` from ``str`` to ``numpy.ndarray``
        :param x_column: Indicates which column should be used as the x parameter
        :type x_column: ``str`` or ``numpy.ndarray``
        :param xerr_column: Indicates which column should be used as the x error
         parameter
        :type xerr_column: ``str`` or ``numpy.ndarray``
        :param y_column: Indicates which column should be used as the y parameter
        :type y_column: ``str`` or ``numpy.ndarray``
        :param yerr_column: Indicates which column should be used as the y error
         parameter
        :type yerr_column: ``str`` or ``numpy.ndarray``
        """
        self._data = OrderedDict(
            [(key, np.array(value)) for key, value in data.items()]
        )
        self._all_columns = list(self.data.keys())
        lengths = {value.size for value in self.data.values()}
        if len(lengths) != 1:
            raise FittingDataColumnsLengthError()
        self._length = next(iter(lengths))
        self._statistics_map: Dict[str, Optional[Statistics]] = OrderedDict()
        self.select_all_records()
        self.__update_statistics()
        self.x_column = x_column
        self.xerr_column = xerr_column
        self.y_column = y_column
        self.yerr_column = yerr_column

    # Data properties are read-only

    @property
    def length(self):
        """Number of records."""
        return self._length

    @property
    def data(self):
        """Data matrix."""
        return self._data

    @property
    def all_records(self) -> List[List[Any]]:
        """Get all records in data as a list."""
        return [list(record) for record in zip(*self.data.values())]

    @property
    def records(self):
        """Get all selected records in data as a list."""
        return list(
            zip(*[column[self.records_indices] for column in self.data.values()])
        )

    @property
    def all_columns(self) -> List[str]:
        """Columns list."""
        return self._all_columns

    @property
    def used_columns(self):
        """Dictionary of columns in use."""
        return Columns(
            x=self.x_column,
            xerr=self.xerr_column,
            y=self.y_column,
            yerr=self.yerr_column,
        )

    def column_data(self, column_header: str):
        """
        Get the data of a column.

        :param column_header: The header name of the desired column.
        :type column_header: str
        :returns: The data of the given column
        :rtype: numpy.array
        """
        return self.data[column_header][self.records_indices]

    @property
    def x(self):  # pylint: disable=invalid-name
        """X values."""
        return self.column_data(self.x_column)

    @property
    def xerr(self):
        """X error values."""
        return self.column_data(self.xerr_column)

    @property
    def y(self):  # pylint: disable=invalid-name
        """Y values."""
        return self.column_data(self.y_column)

    @property
    def yerr(self):
        """Y error values."""
        return self.column_data(self.yerr_column)

    # Records indices methods

    def select_record(self, index: int):
        """
        Select a record to be used in fitting.

        :param index: index of the desired record **starting from 1**.
        :type index: int
        """
        self.records_indices[index - 1] = True
        self.__update_statistics()

    def unselect_record(self, index: int):
        """
        Unselect a record to be used in fitting.

        :param index: index of the desired record **starting from 1**.
        :type index: int
        """
        self.records_indices[index - 1] = False
        self.__update_statistics()

    def select_all_records(self):
        """Select all records to be used in fitting."""
        self.records_indices = [True] * self.length

    def unselect_all_records(self):
        """Unselect all records from being used in fitting."""
        self.records_indices = [False] * self.length

    def is_selected(self, index):
        """
        Checks if a record is selected or not.

        :param index: index of the desired record **starting from 1**.
        :type index: int
        :returns: ``bool``
        """
        return self.records_indices[index - 1]

    @property
    def records_indices(self):
        """List of booleans indicating which records are selected."""
        return self._records_indices

    @records_indices.setter
    def records_indices(self, records_indices):
        if len(records_indices) != self.length:
            raise FittingDataColumnsSelectionError(
                f"Should select {self.length} records,"
                f" only {len(records_indices)} selected."
            )
        if not all([isinstance(element, bool) for element in records_indices]):
            raise FittingDataColumnsSelectionError(
                "When setting record indices, all values should be booleans."
            )
        self._records_indices = records_indices
        self.__update_statistics()

    @property
    def x_column(self):
        """Name of the x column."""
        return self._x_column

    @x_column.setter
    def x_column(self, x_column):
        if x_column is None:
            self._x_column_index = 0
        elif x_column in self.all_columns:
            self._x_column_index = self.all_columns.index(x_column)
        else:
            self._x_column_index = self.__covert_to_index(x_column)
        self.__validate_index(self._x_column_index, x_column)
        self._x_column = self.all_columns[self._x_column_index]

    @property
    def xerr_column(self):
        """Name of the x error column."""
        return self._xerr_column

    @xerr_column.setter
    def xerr_column(self, xerr_column):
        if xerr_column is None:
            self._xerr_column_index = self._x_column_index + 1
        elif xerr_column in self.all_columns:
            self._xerr_column_index = self.all_columns.index(xerr_column)
        else:
            self._xerr_column_index = self.__covert_to_index(xerr_column)
        self.__validate_index(self._xerr_column_index, xerr_column)
        self._xerr_column = self.all_columns[self._xerr_column_index]

    @property
    def y_column(self):
        """Name of the y column."""
        return self._y_column

    @y_column.setter
    def y_column(self, y_column):
        if y_column is None:
            self._y_column_index = self._xerr_column_index + 1
        elif y_column in self.all_columns:
            self._y_column_index = self.all_columns.index(y_column)
        else:
            self._y_column_index = self.__covert_to_index(y_column)
        self.__validate_index(self._y_column_index, y_column)
        self._y_column = self.all_columns[self._y_column_index]

    @property
    def yerr_column(self):
        """Name of the y error column."""
        return self._yerr_column

    @yerr_column.setter
    def yerr_column(self, yerr_column):
        if yerr_column is None:
            self._yerr_column_index = self._y_column_index + 1
        elif yerr_column in self.all_columns:
            self._yerr_column_index = self.all_columns.index(yerr_column)
        else:
            self._yerr_column_index = self.__covert_to_index(yerr_column)
        self.__validate_index(self._yerr_column_index, yerr_column)
        self._yerr_column = self.all_columns[self._yerr_column_index]

    def statistics(self, column_name: str) -> Optional[Statistics]:
        """
        Get statistics of the values in a column.

        :param column_name: The column name to get statistics of
        :type column_name: str
        :returns: Statistics of the given column
        :rtype: Statistics
        """
        return self._statistics_map[column_name]

    # More functionalities

    def residuals(self, fit_func, a: np.ndarray):  # pylint: disable=invalid-name
        """
        Creates residuals :class:`FittingData` objects.

        :param fit_func: :class:`FittingFunction` to evaluate with the fit data
        :type fit_func: ``FittingFunction``
        :param a: the parameters of the given fitting function
        :type a: ``numpy.ndarray``
        :returns: residuals :class:`FittingData`
        """
        y_residuals = self.y - fit_func(a, self.x)
        return FittingData(
            data=OrderedDict(
                [
                    (self.x_column, self.x),
                    (self.xerr_column, self.xerr),
                    (self.y_column, y_residuals),
                    (self.yerr_column, self.yerr),
                ]
            )
        )

    # Read and generate methods

    @classmethod
    def random(  # pylint: disable=invalid-name,too-many-arguments
        cls,
        fit_func,  # type: ignore
        x: Optional[np.ndarray] = None,
        a: Optional[np.ndarray] = None,
        xmin: float = DEFAULT_XMIN,
        xmax: float = DEFAULT_XMAX,
        min_coeff: float = DEFAULT_MIN_COEFF,
        max_coeff: float = DEFAULT_MAX_COEFF,
        xsigma: float = DEFAULT_XSIGMA,
        ysigma: float = DEFAULT_YSIGMA,
        measurements: int = DEFAULT_MEASUREMENTS,
    ):
        """
        Generate a random fit data.

        :param fit_func: :class:`FittingFunction` to evaluate with the fit data
        :type fit_func: ``FittingFunction``
        :param x: Optional. The input for the fitting algorithm.
         If not given, generated randomly.
        :type x: ``numpy.ndarray``
        :param a: Optional. the actual parameters that should be returned by the
         fitting algorithm. If not given, generated randomly.
        :type a: ``numpy.ndarray``
        :param xmin: Minimum value for x.
        :type xmin: float
        :param xmax: Maximum value for x.
        :type xmax: float
        :param min_coeff: Minimum value for `a` coefficient.
        :type min_coeff: float
        :param max_coeff: Maximum value for `a` coefficient.
        :type max_coeff: float
        :param xsigma: Standard deviation for x.
        :type xsigma: int
        :param ysigma: Standard deviation for y.
        :type ysigma: int
        :param measurements: Number of measurements
        :type measurements: int
        :returns: random :class:`FittingData`
        """
        if a is None:
            a = random_array(min_val=min_coeff, max_val=max_coeff, size=fit_func.n)
        if x is None:
            x = random_array(min_val=xmin, max_val=xmax, size=measurements)
        xerr = random_sigma(average_sigma=xsigma, size=measurements)
        yerr = random_sigma(average_sigma=ysigma, size=measurements)
        y = fit_func(a, x + random_error(scales=xerr)) + random_error(scales=yerr)
        return FittingData(
            data=OrderedDict([("x", x), ("xerr", xerr), ("y", y), ("yerr", yerr)])
        )

    @classmethod
    def read_from_excel(  # pylint: disable=too-many-arguments
        cls,
        filepath: Union[str, Path],
        sheet: str,
        x_column: Optional[Union[str, int]] = None,
        xerr_column: Optional[Union[str, int]] = None,
        y_column: Optional[Union[str, int]] = None,
        yerr_column: Optional[Union[str, int]] = None,
    ):
        """
        Read :class:`FittingData` from excel file.

        :param filepath: str or Path. Path to location of excel file
        :param sheet: str. The name of the sheet to extract the data from.
        :param x_column: Indicates which column should be used as the
         x parameter
        :type x_column: ``str`` or ``numpy.ndarray``
        :param xerr_column: Indicates which column should be used as the x error
         parameter
        :type xerr_column: ``str`` or ``numpy.ndarray``
        :param y_column: Indicates which column should be used as the x parameter
        :type y_column: ``str`` or ``numpy.ndarray``
        :param yerr_column: Indicates which column should be used as the y error
         parameter
        :type xerr_column: ``str`` or ``numpy.ndarray``
        :returns: :class:`FittingData` read from the excel file.
        """
        if isinstance(filepath, str):
            filepath = Path(filepath)

        workbook = openpyxl.load_workbook(filepath, data_only=True)
        rows = [list(row) for row in workbook[sheet].values]
        return cls.__build_from_rows(
            rows=rows,
            x_column=x_column,
            xerr_column=xerr_column,
            y_column=y_column,
            yerr_column=yerr_column,
        )

    @classmethod
    def read_from_csv(  # pylint: disable=too-many-arguments
        cls,
        filepath: Union[str, Path],
        x_column: Optional[Union[str, int]] = None,
        xerr_column: Optional[Union[str, int]] = None,
        y_column: Optional[Union[str, int]] = None,
        yerr_column: Optional[Union[str, int]] = None,
    ):
        """
        Read :class:`FittingData` from csv file.

        :param filepath: str or Path. Path to location of csv file
        :param x_column: Indicates which column should be used as the x parameter
        :type x_column: ``str`` or ``numpy.ndarray``
        :param xerr_column: Indicates which column should be used as the x error
         parameter
        :type xerr_column: ``str`` or ``numpy.ndarray``
        :param y_column: Indicates which column should be used as the x parameter
        :type y_column: ``str`` or ``numpy.ndarray``
        :param yerr_column: Indicates which column should be used as the y error
         parameter
        :type xerr_column: ``str`` or ``numpy.ndarray``
        :returns: :class:`FittingData` read from the csv file.
        """
        if isinstance(filepath, str):
            filepath = Path(filepath)
        with open(filepath, mode="r") as csv_file:
            csv_obj = csv.reader(csv_file)
            rows = list(csv_obj)
        return cls.__build_from_rows(
            rows=rows,
            x_column=x_column,
            xerr_column=xerr_column,
            y_column=y_column,
            yerr_column=yerr_column,
        )

    @classmethod
    def read_from_json(  # pylint: disable=too-many-arguments
        cls,
        filepath: Union[str, Path],
        x_column: Optional[Union[str, int]] = None,
        xerr_column: Optional[Union[str, int]] = None,
        y_column: Optional[Union[str, int]] = None,
        yerr_column: Optional[Union[str, int]] = None,
    ):
        """
        Read :class:`FittingData` from json file.

        :param filepath: str or Path. Path to location of csv file
        :param x_column: Indicates which column should be used as the x parameter
        :type x_column: ``str`` or ``numpy.ndarray``
        :param xerr_column: Indicates which column should be used as the x error
         parameter
        :type xerr_column: ``str`` or ``numpy.ndarray``
        :param y_column: Indicates which column should be used as the x parameter
        :type y_column: ``str`` or ``numpy.ndarray``
        :param yerr_column: Indicates which column should be used as the y error
         parameter
        :type xerr_column: ``str`` or ``numpy.ndarray``
        :returns: :class:`FittingData` read from the json file.
        """
        if isinstance(filepath, str):
            filepath = Path(filepath)
        with open(filepath, mode="r") as file:
            data = json.load(file, object_pairs_hook=OrderedDict)
        try:
            # fmt: off
            return FittingData(
                OrderedDict(
                    [(key, list(map(float, row))) for key, row in data.items()]
                ),
                x_column=x_column, xerr_column=xerr_column,
                y_column=y_column, yerr_column=yerr_column,
            )
            # fmt: on
        except (ValueError, TypeError) as error:
            raise FittingDataInvalidFile(
                f'"{filepath.name}" has invalid syntax.'
            ) from error

    # Set methods

    def set_header(self, old, new):
        """
        Rename header.

        :param old: The old columns name
        :type old: str
        :param new: The new value to set for the header
        :type new: str
        """
        if new == old:
            return
        if new == "":
            raise FittingDataSetError("Cannot set new header to be empty")
        if new in self.all_columns:
            raise FittingDataSetError(f'The column name:"{new}" is already used.')
        self._data[new] = self._data.pop(old)
        self._all_columns = list(self.data.keys())
        self.__update_statistics()

    def set_cell(self, record_number, column_name, value):
        """
        Set new value to a cell.

        :param record_number: The number of the record to set, starting from 1
        :type record_number: int
        :param column_name: The column name
        :type column_name: str
        :param value: The new value to set for the cell
        :type value: float
        """
        if not self.__is_number(value):
            raise FittingDataSetError(
                f'The cell at record number:"{record_number}", '
                f'column:"{column_name}" has invalid syntax: {value}.'
            )
        try:
            self._data[column_name][record_number - 1] = value
        except KeyError as error:
            raise FittingDataSetError(
                f'Column name "{column_name}" does not exists'
            ) from error
        except IndexError as error:
            raise FittingDataSetError(
                f"Record number {record_number} does not exists"
            ) from error
        self.__update_statistics()

    # Save methods

    def save_excel(
        self,
        output_directory: Union[str, Path],
        name: str = "fitting_data",
        sheet: Optional[str] = None,
    ):
        """
        Save :class:`FittingData` to xlsx file.

        :param output_directory: Path to the directory for the new excel file to be
         saved.
        :type output_directory: ``Path`` or ``str``
        :param name: Optional. The name of the file, without the .xlsx suffix.
         "fitting_data" by default.
        :type name: str
        :param sheet: Optional. Name of the sheet that the data will be saved to.
        :type sheet: str
        """
        io_util.save_as_excel(
            content=[self.all_columns, *self.all_records],
            output_directory=output_directory,
            file_name=name,
            sheet=sheet,
        )

    def save_csv(self, output_directory: Union[str, Path], name: str = "fitting_data"):
        """
        Save :class:`FittingData` to csv file.

        :param output_directory:
         Path to the directory for the new excel file to be saved.
        :type output_directory: ``Path`` or ``str``
        :param name: Optional. The name of the file, without the .csv suffix.
         "fitting_data" by default.
        :type name: str
        """
        io_util.save_as_csv(
            content=[self.all_columns, *self.all_records],
            output_directory=output_directory,
            file_name=name,
        )

    def save_statistics_excel(
        self,
        output_directory: Union[str, Path],
        name: Optional[str] = None,
        sheet: Optional[str] = None,
    ):
        """
        Save the fitting data statistics to xlsx file.

        :param output_directory: Path to the directory for the new excel file to be
         saved.
        :type output_directory: ``Path`` or ``str``
        :param name: Optional. The name of the file, without the .xlsx suffix.
         "fitting_data_statistics" by default.
        :type name: str
        :param sheet: Optional. Name of the sheet that the data will be saved to.
        :type sheet: str
        """
        if name is None:
            name = "fitting_data_statistics"
        io_util.save_as_excel(
            content=self.__statistics_as_records(),
            output_directory=output_directory,
            file_name=name,
            sheet=sheet,
        )

    def save_statistics_csv(
        self, output_directory: Union[str, Path], name: Optional[str] = None
    ):
        """
        Save the fitting data statistics to csv file.

        :param output_directory:
         Path to the directory for the new excel file to be saved.
        :type output_directory: ``Path`` or ``str``
        :param name: Optional. The name of the file, without the .csv suffix.
         "fitting_data" by default.
        :type name: str
        """
        if name is None:
            name = "fitting_data_statistics"
        io_util.save_as_csv(
            content=self.__statistics_as_records(),
            output_directory=output_directory,
            file_name=name,
        )

    # Build from rows

    @classmethod
    def __build_from_rows(  # pylint: disable=too-many-arguments
        cls,
        rows,
        x_column: Optional[Union[str, int]] = None,
        xerr_column: Optional[Union[str, int]] = None,
        y_column: Optional[Union[str, int]] = None,
        yerr_column: Optional[Union[str, int]] = None,
    ):
        rows = cls.__trim_data(rows)
        if len(rows) == 0:
            raise FittingDataInvalidFile("All rows are empty.")
        headers, content = cls.__extract_headers(rows)
        cls.__validate_headers(headers)
        data = cls.__build_raw_data(headers, content)
        return FittingData(
            data=data,
            x_column=x_column,
            xerr_column=xerr_column,
            y_column=y_column,
            yerr_column=yerr_column,
        )

    @classmethod
    def __trim_data(cls, rows):
        if len(rows) == 0:
            return rows
        first_row = rows[0]
        row_length = None
        for i, val in enumerate(first_row):
            if cls.__is_empty_value(val):
                row_length = i
                break
        if row_length is None:
            row_length = len(first_row)
        new_rows = []
        for i, row in enumerate(rows):
            row = list(row)
            if len(row) > row_length:
                if not cls.__is_empty_value(row[row_length]):
                    raise FittingDataInvalidFile(
                        f"Cell should be empty at row {i} column {row_length + 1}."
                    )
                row = row[:row_length]
            while len(row) != 0 and cls.__is_empty_value(row[-1]):
                row = row[:-1]
            if len(row) == 0:
                break
            if len(row) < row_length:
                row.extend([None for _ in range(row_length - len(row))])
            new_rows.append(row)
        return new_rows

    @classmethod
    def __extract_headers(cls, rows: List[List[Optional[str]]]):
        headers: List[Optional[str]] = rows[0]
        if cls.__are_headers(headers):
            content = rows[1:]
        else:
            headers = [str(i) for i in range(len(headers))]
            content = rows
        return headers, content

    @classmethod
    def __validate_headers(cls, headers):
        duplicate_headers = [
            item for item, count in collections.Counter(headers).items() if count > 1
        ]
        if len(duplicate_headers) != 0:
            raise FittingDataInvalidFile(
                f"The following headers appear more than once: "
                f'{", ".join(duplicate_headers)}'
            )

    @classmethod
    def __build_raw_data(cls, headers, content):
        content = [cls.__build_row(i, row) for i, row in enumerate(content, start=1)]
        columns = [np.array(column) for column in zip(*content)]
        return collections.OrderedDict(zip(headers, columns))

    @classmethod
    def __build_row(cls, row_number, row):
        return [
            cls.__build_cell(row_number, column_number, val)
            for column_number, val in enumerate(row, start=1)
        ]

    @classmethod
    def __build_cell(cls, row_number, column_number, val):
        if isinstance(val, str):
            val = val.strip()
        if cls.__is_empty_value(val):
            raise FittingDataInvalidFile(
                f"Empty cell at column {column_number} row {row_number}."
            )
        try:
            return float(val)
        except ValueError as error:
            raise FittingDataInvalidFile(
                f"Cell should be a number at column {column_number} row {row_number}, "
                f'got "{val}".'
            ) from error

    @classmethod
    def __is_empty_value(cls, val: Optional[Union[str, float]]):
        if val is None:
            return True
        if isinstance(val, str) and val.strip() == "":
            return True
        return False

    @classmethod
    def __are_headers(cls, headers):
        return all([cls.__is_header(header) for header in headers])

    @classmethod
    def __is_header(cls, string):
        return isinstance(string, str) and not cls.__is_number(string)

    @classmethod
    def __is_number(cls, string):
        try:
            float(string)
            return True
        except ValueError:
            return False

    # Private methods

    def __validate_index(self, index, column):
        if index is None:
            raise FittingDataColumnExistenceError(column)
        max_index = len(self._all_columns)
        if index < 0 or index >= max_index:
            raise FittingDataColumnIndexError(index + 1, max_index)

    def __update_statistics(self):
        self._statistics_map.clear()
        for column in self.all_columns:
            try:
                self._statistics_map[column] = Statistics.from_array(
                    self.column_data(column)
                )
            except ValueError:
                self._statistics_map[column] = None

    def __statistics_as_records(self):
        records = [["Parameters", *self.all_columns]]
        for parameter in Statistics.parameters():
            records.append(
                [
                    parameter.replace("_", " ").title(),
                    *[
                        getattr(statistics, parameter)
                        for statistics in self._statistics_map.values()
                    ],
                ]
            )
        return records

    @classmethod
    def __covert_to_index(cls, column):
        try:
            return int(column) - 1
        except ValueError:
            return None



# pylint: disable=missing-class-docstring
"""Exception classes for Eddington."""


class EddingtonException(Exception):  # noqa: D101
    pass


# Fitting Function Errors


class FittingFunctionLoadError(EddingtonException):  # noqa: D101
    pass


class FittingFunctionSaveError(EddingtonException):  # noqa: D101
    pass


class FittingFunctionRuntimeError(EddingtonException):  # noqa: D101
    pass


# Fitting Data Errors


class FittingDataError(EddingtonException):  # noqa: D101
    pass


class FittingDataColumnsLengthError(FittingDataError):  # noqa: D101
    msg = "All columns in FittingData should have the same length"

    def __init__(self) -> None:  # noqa: D107
        super().__init__(self.msg)


class FittingDataInvalidFile(FittingDataError):  # noqa: D101
    pass


class FittingDataColumnIndexError(FittingDataError):  # noqa: D101
    def __init__(self, index: int, max_index: int) -> None:  # noqa: D107
        super().__init__(
            f"No column number {index} in data. "
            f"index should be between 1 and {max_index}"
        )


class FittingDataColumnExistenceError(FittingDataError):  # noqa: D101
    def __init__(self, column: str) -> None:  # noqa: D107
        super().__init__(f'Could not find column "{column}" in data')


class FittingDataColumnsSelectionError(FittingDataError):  # noqa: D101
    pass


class FittingDataSetError(FittingDataError):  # noqa: D101
    pass


# Plot Errors


class PlottingError(EddingtonException):  # noqa: D101
    pass



"""Plotting methods."""
from typing import Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

from eddington.exceptions import PlottingError
from eddington.fitting_data import FittingData
from eddington.print_util import to_relevant_precision_string


class Figure:
    """
    Wraps matplotlib Figure class.

    It releases the memory when the figure is not longer in use.
    """

    def __init__(self, fig):
        """Figure constructor."""
        self._actual_fig = fig

    def __enter__(self):
        """Return self when entering as context."""
        return self

    def __getattr__(self, item):
        """Get attributes from wrapped figure."""
        return getattr(self._actual_fig, item)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Clear memory on exit."""
        plt.clf()
        plt.close("all")


def plot_residuals(  # pylint: disable=invalid-name,too-many-arguments
    func,
    data: FittingData,
    a: np.ndarray,
    title_name,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    grid: bool = False,
    x_log_scale: bool = False,
    y_log_scale: bool = False,
    xmin: Optional[float] = None,
    xmax: Optional[float] = None,
):
    """
    Plot residuals plot.

    :param func: Fitting function.
    :type func: :class:`FittingFunction`
    :param data: Fitting data
    :type data: :class:`FittingData`
    :param a: The parameters result
    :type a: ``numpy.ndarray`` or ``list``
    :param title_name: Optional. Title for the figure.
    :type title_name: str
    :param xlabel: Optional. Label of the x axis
    :type xlabel: str
    :param ylabel: Optional. Label of the x axis
    :type ylabel: str
    :param grid: Add grid lines or not
    :type grid: bool
    :param x_log_scale: Set the scale of the  x axis to be logarithmic.
    :type x_log_scale: bool
    :param y_log_scale: Set the scale of the y axis to be logarithmic.
    :type y_log_scale: bool
    :param xmin: Optional. minimum value for x in plot
    :type xmin: float
    :param xmax: Optional. maximum value for x in plot
    :type xmax: float
    :returns: ``matplotlib.pyplot.Figure``
    """
    ax, fig = get_figure(
        title_name=title_name,
        xlabel=xlabel,
        ylabel=ylabel,
        grid=grid,
        x_log_scale=x_log_scale,
        y_log_scale=y_log_scale,
    )
    xmin, xmax = get_plot_borders(x=data.x, xmin=xmin, xmax=xmax)
    errorbar(ax=ax, data=data.residuals(func, a), xmin=xmin, xmax=xmax)
    horizontal_line(ax=ax, xmin=xmin, xmax=xmax)
    return fig


def plot_fitting(  # pylint: disable=C0103,R0913,R0914
    func,
    data: FittingData,
    a: Union[np.ndarray, List[np.ndarray], Dict[str, np.ndarray]],
    title_name,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    grid: bool = False,
    legend: Optional[bool] = None,
    x_log_scale: bool = False,
    y_log_scale: bool = False,
    step: Optional[float] = None,
    xmin: Optional[float] = None,
    xmax: Optional[float] = None,
):
    """
    Plot fitting plot.

    :param func: Fitting function.
    :type func: :class:`FittingFunction`
    :param data: Fitting data
    :type data: :class:`FittingData`
    :param a: The parameters result
    :type a: ``numpy.ndarray``, a list of ``numpy.ndarray`` items or a dictionary from
     strings to ``numpy.ndarray``
    :param title_name: Optional. Title for the figure.
    :type title_name: str
    :param xlabel: Optional. Label of the x axis
    :type xlabel: str
    :param ylabel: Optional. Label of the x axis
    :type ylabel: str
    :param grid: Add grid lines or not
    :type grid: bool
    :param legend: Optional. Add legend or not. If None, add legend when more than
     one parameters values has been presented.
    :type legend: bool
    :param x_log_scale: Set the scale of the  x axis to be logarithmic.
    :type x_log_scale: bool
    :param y_log_scale: Set the scale of the y axis to be logarithmic.
    :type y_log_scale: bool
    :param step: Optional. Steps between samples for the fitting graph
    :type step: float
    :param xmin: Optional. minimum value for x in plot
    :type xmin: float
    :param xmax: Optional. maximum value for x in plot
    :type xmax: float
    :returns: ``matplotlib.pyplot.Figure``
    """
    ax, fig = get_figure(
        title_name=title_name,
        xlabel=xlabel,
        ylabel=ylabel,
        grid=grid,
        x_log_scale=x_log_scale,
        y_log_scale=y_log_scale,
    )
    xmin, xmax = get_plot_borders(x=data.x, xmin=xmin, xmax=xmax)
    errorbar(ax=ax, data=data, xmin=xmin, xmax=xmax)
    if step is None:
        step = (xmax - xmin) / 1000.0
    x = np.arange(xmin, xmax, step=step)  # pylint: disable=invalid-name
    a_dict = __get_a_dict(a)
    for label, a_value in a_dict.items():
        add_plot(ax=ax, x=x, y=func(a_value, x), label=label)
    add_legend(ax, __should_add_legend(legend, a_dict))
    return fig


def plot_data(  # pylint: disable=too-many-arguments
    data: FittingData,
    title_name,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    xmin: Optional[float] = None,
    xmax: Optional[float] = None,
    grid: bool = False,
    x_log_scale: bool = False,
    y_log_scale: bool = False,
):
    """
    Plot fitting data.

    :param data: Fitting data
    :type data: :class:`FittingData`
    :param title_name: Optional. Title for the figure.
    :type title_name: str
    :param xlabel: Optional. Label of the x axis
    :type xlabel: str
    :param ylabel: Optional. Label of the x axis
    :type ylabel: str
    :param xmin: Optional. Minimum value for x. if None, calculated from given data
    :type xmin: float
    :param xmax: Optional. Maximum value for x. if None, calculated from given data
    :type xmax: float
    :param grid: Add grid lines or not
    :type grid: bool
    :param x_log_scale: Set the scale of the  x axis to be logarithmic.
    :type x_log_scale: bool
    :param y_log_scale: Set the scale of the y axis to be logarithmic.
    :type y_log_scale: bool
    :returns: ``matplotlib.pyplot.Figure``
    """
    ax, fig = get_figure(  # pylint: disable=invalid-name
        title_name=title_name,
        xlabel=xlabel,
        ylabel=ylabel,
        grid=grid,
        x_log_scale=x_log_scale,
        y_log_scale=y_log_scale,
    )
    xmin, xmax = get_plot_borders(x=data.x, xmin=xmin, xmax=xmax)
    errorbar(ax=ax, data=data, xmin=xmin, xmax=xmax)
    limit_axes(ax=ax, xmin=xmin, xmax=xmax)
    return fig


def get_figure(  # pylint: disable=too-many-arguments
    title_name,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    grid: bool = False,
    x_log_scale: bool = False,
    y_log_scale: bool = False,
):
    """
    Gets a figure from matplotlib.

    :param title_name: Optional. Title for the figure.
    :type title_name: str
    :param xlabel: Optional. Label of the x axis
    :type xlabel: str
    :param ylabel: Optional. Label of the x axis
    :type ylabel: str
    :param grid: Add grid lines or not
    :type grid: bool
    :param x_log_scale: Set the scale of the  x axis to be logarithmic.
    :type x_log_scale: bool
    :param y_log_scale: Set the scale of the y axis to be logarithmic.
    :type y_log_scale: bool
    :return: Figure instance
    """
    fig = plt.figure()
    ax = fig.add_subplot()  # pylint: disable=invalid-name
    title(ax=ax, title_name=title_name)
    label_axes(ax=ax, xlabel=xlabel, ylabel=ylabel)
    add_grid(ax=ax, is_grid=grid)
    set_scales(ax=ax, is_x_log_scale=x_log_scale, is_y_log_scale=y_log_scale)

    return ax, Figure(fig)


def title(ax: plt.Axes, title_name: Optional[str]):  # pylint: disable=invalid-name
    """
    Add/remove title to figure.

    :param ax: Figure axes.
    :type ax: matplotlib.pyplot.Axes
    :param title_name: Optional. If None, don't add title. otherwise, add given title
    :type title_name: str
    """
    if title_name is not None:
        ax.set_title(title_name)


def label_axes(  # pylint: disable=invalid-name
    ax: plt.Axes, xlabel: Optional[str], ylabel: Optional[str]
):
    """
    Add/remove labels to figure.

    :param ax: Figure axes.
    :type ax: matplotlib.pyplot.Axes
    :param xlabel: Optional. If None, don't add label. otherwise, add given label
    :type xlabel: str
    :param ylabel: Optional. If None, don't add label. otherwise, add given label
    :type ylabel: str
    """
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)


def limit_axes(  # pylint: disable=invalid-name
    ax: plt.Axes, xmin: Optional[float] = None, xmax: Optional[float] = None
):
    """
    Set limits on axes.

    :param ax: Figure axes.
    :type ax: matplotlib.pyplot.Axes
    :param xmin: Optional. Minimum value for x. if None, calculated from given data
    :type xmin: float
    :param xmax: Optional. Maximum value for x. if None, calculated from given data
    :type xmax: float
    """
    if xmin is not None:
        ax.set_xlim(left=xmin)
    if xmax is not None:
        ax.set_xlim(right=xmax)


def add_grid(ax: plt.Axes, is_grid: bool):  # pylint: disable=invalid-name
    """
    Add/remove grid to figure.

    :param ax: Figure axes.
    :type ax: matplotlib.pyplot.Axes
    :param is_grid: Add or remote grid to plot
    :type is_grid: bool
    """
    ax.grid(is_grid)


def add_legend(ax: plt.Axes, is_legend: bool):  # pylint: disable=invalid-name
    """
    Add/remove legend to figure.

    :param ax: Figure axes.
    :type ax: matplotlib.pyplot.Axes
    :param is_legend: Add or remote legend to plot
    :type is_legend: bool
    """
    if is_legend:
        ax.legend()


def set_scales(  # pylint: disable=invalid-name
    ax: plt.Axes, is_x_log_scale: bool, is_y_log_scale: bool
):
    """
    Change x axis scale to logarithmic.

    :param ax: Figure axes
    :type ax: matplotlib.pyplot.Axes
    :param is_x_log_scale: Change x axis scale to logarithmic or not.
    :type is_x_log_scale: bool
    :param is_y_log_scale: Change y axis scale to logarithmic or not.
    :type is_y_log_scale: bool
    """
    if is_x_log_scale:
        ax.set_xscale("log")
    if is_y_log_scale:
        ax.set_yscale("log")


def add_plot(
    ax: plt.Axes,
    x: Union[np.ndarray, List[float]],
    y: Union[np.ndarray, List[float]],
    label: Optional[str] = None,
):  # pylint: disable=C0103
    """
    Plot y as a function of x.

    :param ax: Figure axes.
    :type ax: matplotlib.pyplot.Axes
    :param x: X values
    :type x: list of floats or ``numpy.ndarray``
    :param y: Y values
    :type y: list of floats or ``numpy.ndarray``
    :param label: Optional. Label for the plot that would be added to the legend
    :type label: str
    """
    ax.plot(x, y, label=label)


def horizontal_line(  # pylint: disable=C0103
    ax: plt.Axes, xmin: float, xmax: float, y: float = 0
):
    """
    Add horizontal line to figure.

    :param ax: Figure axes.
    :type ax: matplotlib.pyplot.Axes
    :param xmin: Minimum x value of line
    :type xmin: float
    :param xmax: Maximum x value of line
    :type xmax: float
    :param y: The y value of the line
    :type y: float
    """
    ax.hlines(y, xmin=xmin, xmax=xmax, linestyles="dashed")


def errorbar(  # pylint: disable=invalid-name
    ax: plt.Axes,
    data: FittingData,
    xmin: Optional[float] = None,
    xmax: Optional[float] = None,
):
    """
    Plot error bar to figure.

    :param ax: Figure axes.
    :type ax: matplotlib.pyplot.Axes
    :param data: Data to visualize
    :type data: eddington.fitting_data.FittingData
    :param xmin: Optional. Minimum value of X. Used for filtering data points
    :type xmin: Optional float
    :param xmax: Optional. Maximum value of X. Used for filtering data points
    :type xmax: Optional float
    """
    checkers_list = [__in_bounds(val, min_val=xmin, max_val=xmax) for val in data.x]
    ax.errorbar(
        x=data.x[checkers_list],
        y=data.y[checkers_list],
        xerr=data.xerr[checkers_list],
        yerr=data.yerr[checkers_list],
        markersize=1,
        marker="o",
        linestyle="None",
    )


def get_plot_borders(  # pylint: disable=invalid-name
    x: np.ndarray, xmin: Optional[float] = None, xmax: Optional[float] = None
) -> Tuple[float, float]:
    """
    Get borders for a plot based on its x values.

    :param x: x values list or array.
    :type x: ``numpy.ndarray`` or ``list``
    :param xmin: Minimum x value of line
    :type xmin: float
    :param xmax: Maximum x value of line
    :type xmax: float
    :return: tuple. minimum and maximum values for the plot.
    """
    data_xmin = np.min(x)
    data_xmax = np.max(x)
    gap = (data_xmax - data_xmin) * 0.1
    if xmin is None:
        xmin = data_xmin - gap
    if xmax is None:
        xmax = data_xmax + gap
    return xmin, xmax


def show_or_export(fig: plt.Figure, output_path=None):
    """
    Show plot or export it to a file.

    :param fig: a plot figure
    :param output_path: Path or None. if None, show plot. otherwise, save to path.
    """
    if output_path is None:
        plt.show()
        return
    fig.savefig(output_path)


def __get_a_dict(a):  # pylint: disable=invalid-name
    if isinstance(a, dict):
        return a
    if isinstance(a, list):
        return {__build_repr_string(a_value): a_value for a_value in a}
    if isinstance(a, np.ndarray):
        return {__build_repr_string(a): a}
    raise PlottingError(
        f"{a} has unmatching type. Can except only numpy arrays, "
        "lists of numpy arrays and dictionaries."
    )


def __build_repr_string(a):  # pylint: disable=invalid-name
    arguments_values = [
        f"a[{i}]={to_relevant_precision_string(val)}" for i, val in enumerate(a)
    ]
    return f"[{', '.join(arguments_values)}]"


def __should_add_legend(legend, a_dict):  # pylint: disable=invalid-name
    if legend is not None:
        return legend
    if len(a_dict) >= 2:
        return True
    return False


def __in_bounds(val, min_val=None, max_val=None):  # pylint: disable=invalid-name
    if min_val is not None and val < min_val:
        return False
    if max_val is not None and val > max_val:
        return False
    return True



"""list CLI method."""
import re
from typing import Optional

import click
from prettytable import PrettyTable

from eddington import FittingFunctionsRegistry
from eddington.cli.main_cli import eddington_cli


@eddington_cli.command("list")
@click.option(
    "-r",
    "--regex",
    type=str,
    default=None,
    help="Filter functions by a regular expression",
)
def eddington_list(regex: Optional[str]):
    """Prints all fitting functions in a pretty table."""
    table = PrettyTable(field_names=["Function", "Syntax"])
    for func in FittingFunctionsRegistry.all():
        if regex is None or re.search(regex, func.name):
            table.add_row([func.name, func.syntax])
    click.echo(table)



"""Module for saving content."""
import csv
from pathlib import Path
from typing import Any, List, Optional, Union

import numpy as np
import openpyxl


def save_as_excel(
    content: List[Union[List[Any], np.ndarray]],
    output_directory: Union[str, Path],
    file_name: str,
    sheet: Optional[str] = None,
):
    """
    Save content to xlsx file.

    :param content: list of list, each represent a row in the excel file
    :type content: List[List[Any]] or numpy.ndarray
    :param output_directory: Path to the directory for the new excel file to be
     saved.
    :type output_directory: ``Path`` or ``str``
    :param file_name: The name of the file without suffix.
    :type file_name: str
    :param sheet: Optional. Name of the sheet that the data will be saved to.
    :type sheet: str or None
    """
    workbook = openpyxl.Workbook()
    worksheet = workbook.active

    if sheet:
        worksheet.title = sheet

    for row in content:
        worksheet.append(row)

    path = Path(output_directory) / f"{file_name}.xlsx"

    workbook.save(path)


def save_as_csv(
    content: List[Union[List[Any], np.ndarray]],
    file_name: str,
    output_directory: Union[str, Path],
):
    """
    Save content to csv file.

    :param content: list of list, each represent a row in the excel file
    :type content: List[List[Any]] or numpy.ndarray
    :param file_name: The name of the file without suffix.
    :type file_name: str
    :param output_directory:
     Path to the directory for the new excel file to be saved.
    :type output_directory: ``Path`` or ``str``
    """
    path = Path(output_directory / Path(f"{file_name}.csv"))

    with open(path, mode="w+", newline="") as csv_file:
        writer = csv.writer(csv_file)
        writer.writerows(content)



"""Core functionalities of the Eddington platform."""
from eddington.exceptions import (
    EddingtonException,
    FittingDataColumnExistenceError,
    FittingDataColumnIndexError,
    FittingDataColumnsLengthError,
    FittingDataError,
    FittingDataInvalidFile,
    FittingFunctionLoadError,
    FittingFunctionRuntimeError,
)
from eddington.fitting import fit
from eddington.fitting_data import FittingData
from eddington.fitting_function_class import FittingFunction, fitting_function
from eddington.fitting_functions_list import (
    constant,
    cos,
    exponential,
    hyperbolic,
    inverse_power,
    linear,
    normal,
    parabolic,
    poisson,
    polynomial,
    sin,
    straight_power,
)
from eddington.fitting_functions_registry import FittingFunctionsRegistry
from eddington.fitting_result import FittingResult
from eddington.plot import (
    add_grid,
    add_legend,
    add_plot,
    errorbar,
    get_figure,
    horizontal_line,
    label_axes,
    plot_data,
    plot_fitting,
    plot_residuals,
    show_or_export,
    title,
)
from eddington.print_util import to_relevant_precision_string

__version__ = "0.0.23"

__all__ = [
    "__version__",
    # Fitting functions infrastructure
    "FittingFunction",
    "fitting_function",
    "FittingFunctionsRegistry",
    # Fitting functions
    "constant",
    "exponential",
    "hyperbolic",
    "linear",
    "parabolic",
    "polynomial",
    "cos",
    "sin",
    "straight_power",
    "inverse_power",
    "normal",
    "poisson",
    # Fitting algorithm
    "fit",
    # Exceptions
    "EddingtonException",
    "FittingFunctionRuntimeError",
    "FittingFunctionLoadError",
    "FittingDataError",
    "FittingDataColumnExistenceError",
    "FittingDataColumnIndexError",
    "FittingDataInvalidFile",
    "FittingDataColumnsLengthError",
    # Data structures
    "FittingData",
    "FittingResult",
    # Plot
    "plot_data",
    "plot_fitting",
    "plot_residuals",
    "show_or_export",
    "get_figure",
    "errorbar",
    "add_legend",
    "add_grid",
    "add_plot",
    "horizontal_line",
    "label_axes",
    "title",
    # Print
    "to_relevant_precision_string",
]



"""List of common fitting functions."""
from typing import Union

import numpy as np
import scipy.special

from eddington.exceptions import FittingFunctionLoadError
from eddington.fitting_function_class import FittingFunction, fitting_function


@fitting_function(
    n=2,
    syntax="a[0] + a[1] * x",
    x_derivative=lambda a, x: np.full(shape=np.shape(x), fill_value=a[1]),
    a_derivative=lambda a, x: np.stack([np.ones(shape=np.shape(x)), x]),
)  # pylint: disable=C0103
def linear(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Simple linear fitting function."""
    return a[0] + a[1] * x


@fitting_function(
    n=1,
    syntax="a[0]",
    x_derivative=lambda a, x: np.zeros(shape=np.shape(x)),
    a_derivative=lambda a, x: np.stack([np.ones(shape=np.shape(x))]),
)  # pylint: disable=C0103
def constant(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Constant fitting function."""
    return np.full(fill_value=a[0], shape=np.shape(x))


@fitting_function(
    n=3,
    syntax="a[0] + a[1] * x + a[2] * x ^ 2",
    x_derivative=lambda a, x: a[1] + 2 * a[2] * x,
    a_derivative=lambda a, x: np.stack([np.ones(shape=np.shape(x)), x, x ** 2]),
)  # pylint: disable=C0103
def parabolic(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Parabolic fitting function."""
    return a[0] + a[1] * x + a[2] * x ** 2


@fitting_function(
    n=4,
    syntax="a[0] * (x + a[1]) ^ a[2] + a[3]",
    x_derivative=lambda a, x: a[2] * a[0] * (x + a[1]) ** (a[2] - 1),
    a_derivative=lambda a, x: np.stack(
        [
            np.power(x + a[1], a[2]),
            a[2] * a[0] * np.power(x + a[1], a[2] - 1),
            a[0] * np.log(x + a[1]) * np.power(x + a[1], a[2]),
            np.ones(shape=np.shape(x)),
        ]
    ),
)  # pylint: disable=C0103
def straight_power(
    a: np.ndarray, x: Union[np.ndarray, float]
) -> Union[np.ndarray, float]:  # pylint: disable=C0103
    """Represent fitting of y ~ x^n."""
    return a[0] * np.power(x + a[1], a[2]) + a[3]


@fitting_function(
    n=4,
    syntax="a[0] / (x + a[1]) ^ a[2] + a[3]",
    x_derivative=lambda a, x: -a[2] * a[0] / np.power(x + a[1], a[2] + 1),
    a_derivative=lambda a, x: np.stack(
        [
            1 / np.power(x + a[1], a[2]),
            -a[2] * a[0] / np.power(x + a[1], a[2] + 1),
            -a[0] * np.log(x + a[1]) * np.power(x + a[1], a[2]),
            np.ones(shape=np.shape(x)),
        ]
    ),
)  # pylint: disable=C0103
def inverse_power(
    a: np.ndarray, x: Union[np.ndarray, float]
) -> Union[np.ndarray, float]:  # pylint: disable=C0103
    """Represent fitting of y ~ x^(-n)."""
    return a[0] / np.power(x + a[1], a[2]) + a[3]


@fitting_function(
    n=3,
    syntax="a[0] / (x + a[1]) + a[2]",
    x_derivative=lambda a, x: -a[0] / ((x + a[1]) ** 2),
    a_derivative=lambda a, x: np.stack(
        [1 / (x + a[1]), -a[0] / ((x + a[1]) ** 2), np.ones(shape=np.shape(x))]
    ),
)  # pylint: disable=C0103
def hyperbolic(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Hyperbolic fitting function."""
    return a[0] / (x + a[1]) + a[2]


@fitting_function(
    n=3,
    syntax="a[0] * exp(a[1] * x) + a[2]",
    x_derivative=lambda a, x: a[0] * a[1] * np.exp(a[1] * x),
    a_derivative=lambda a, x: np.stack(
        [np.exp(a[1] * x), a[0] * x * np.exp(a[1] * x), np.ones(np.shape(x))]
    ),
)  # pylint: disable=C0103
def exponential(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Exponential fitting function."""
    return a[0] * np.exp(a[1] * x) + a[2]


@fitting_function(
    n=4,
    syntax="a[0] * cos(a[1] * x + a[2]) + a[3]",
    x_derivative=lambda a, x: -a[0] * a[1] * np.sin(a[1] * x + a[2]),
    a_derivative=lambda a, x: np.stack(
        [
            np.cos(a[1] * x + a[2]),
            -a[0] * x * np.sin(a[1] * x + a[2]),
            -a[0] * np.sin(a[1] * x + a[2]),
            np.ones(shape=np.shape(x)),
        ]
    ),
)  # pylint: disable=C0103
def cos(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Cosines fitting function."""
    return a[0] * np.cos(a[1] * x + a[2]) + a[3]


@fitting_function(
    n=4,
    syntax="a[0] * sin(a[1] * x + a[2]) + a[3]",
    x_derivative=lambda a, x: a[0] * a[1] * np.cos(a[1] * x + a[2]),
    a_derivative=lambda a, x: np.stack(
        [
            np.sin(a[1] * x + a[2]),
            a[0] * x * np.cos(a[1] * x + a[2]),
            a[0] * np.cos(a[1] * x + a[2]),
            np.ones(shape=np.shape(x)),
        ]
    ),
)  # pylint: disable=C0103
def sin(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Sine fitting function."""
    return a[0] * np.sin(a[1] * x + a[2]) + a[3]


@fitting_function(
    n=4,
    syntax="a[0] * exp( - ((x - a[1]) / a[2]) ^ 2) + a[3]",
    x_derivative=lambda a, x: a[0]
    * np.exp(-(((x - a[1]) / a[2]) ** 2))  # noqa: W503
    * (-2 * (x - a[1]) / a[2]),  # noqa: W503
    a_derivative=lambda a, x: np.stack(
        [
            np.exp(-(((x - a[1]) / a[2]) ** 2)),
            a[0] * np.exp(-(((x - a[1]) / a[2]) ** 2)) * (2 * (x - a[1]) / a[2]),
            a[0] * np.exp(-(((x - a[1]) / a[2]) ** 2)) * (2 * (x - a[1]) / (a[2] ** 2)),
            np.ones(shape=np.shape(x)),
        ]
    ),
)  # pylint: disable=C0103
def normal(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Normal distribution fitting function."""
    return a[0] * np.exp(-(((x - a[1]) / a[2]) ** 2)) + a[3]


@fitting_function(
    n=3,
    syntax="a[0] * (a[1] ^ x) * exp(-a[1]) / gamma(x+1) + a[2]",
    x_derivative=lambda a, x: (
        a[0] * np.power(a[1], x) * np.exp(-a[1]) / scipy.special.gamma(x + 1)
    )  # noqa: W503
    * (np.log(a[1]) - scipy.special.digamma(x + 1)),  # noqa: W503
    a_derivative=lambda a, x: np.stack(
        [
            np.power(a[1], x) * np.exp(-a[1]) / scipy.special.gamma(x + 1),
            (a[0] * np.exp(-a[1]) / scipy.special.gamma(x + 1))
            * (x * np.power(a[1], x - 1) - np.power(a[1], x)),  # noqa: W503
            np.ones(shape=np.shape(x)),
        ]
    ),
)  # pylint: disable=C0103
def poisson(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """Poisson fitting function."""
    return a[0] * np.power(a[1], x) * np.exp(-a[1]) / scipy.special.gamma(x + 1) + a[2]


def polynomial(n: int) -> FittingFunction:  # pylint: disable=C0103
    """
    Creates a polynomial fitting function with parameters as coefficients.

    :param n: Degree of the polynom.
    :return: :class:`FittingFunction`
    """
    n = int(n)
    if n <= 0:
        raise FittingFunctionLoadError(f"n must be positive, got {n}")

    if n == 1:
        return linear

    arange = np.arange(1, n + 1)

    syntax = "a[0] + a[1] * x + " + " + ".join(
        [f"a[{i}] * x ^ {i}" for i in arange[1:]]
    )

    @fitting_function(
        n=n + 1,
        name=f"polynomial_{n}",
        syntax=syntax,
        x_derivative=lambda a, x: polynomial(n - 1)(arange * a[1:], x),
        a_derivative=lambda a, x: np.stack([x ** i for i in range(n + 1)]),
        save=False,
    )  # pylint: disable=C0103
    def func(a: np.ndarray, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
        return sum([a[i] * x ** i for i in range(n + 1)])

    return func



"""Printing functions for displaying numbers with given precision."""
import numpy as np

from eddington.consts import DEFAULT_MAX_STRING_LENGTH, DEFAULT_PRECISION


def to_relevant_precision_string(
    decimal: float,
    relevant_digits: int = DEFAULT_PRECISION,
    max_string_length: int = DEFAULT_MAX_STRING_LENGTH,
) -> str:
    """
    Convert a decimal into string while preserving relevant digits.

    :param decimal: The float to convert to string.
    :type decimal: float
    :param relevant_digits: number of relevant digits to preserve while converting to
        string.
    :type relevant_digits: int
    :param max_string_length: maximum string length when decimal is not normalized.
    :type max_string_length: int
    :returns: decimal as string
    :rtype: str
    """
    if is_nan_or_infinity(decimal):
        return str(decimal)
    order = order_of_magnitude(decimal)
    digit = order - relevant_digits
    return to_digit_string(decimal, digit, max_string_length)


def to_digit_string(
    decimal: float, digit: int, max_string_length: int = DEFAULT_MAX_STRING_LENGTH
) -> str:
    """
    Convert a decimal into string while preserving given digit.

    :param decimal: The float to convert to string.
    :type decimal: float
    :param digit: The lowest digit to preserve while converting to string.
    :type digit: int
    :param max_string_length: maximum string length when decimal is not normalized.
    :type max_string_length: int
    :returns: decimal as string
    :rtype: str
    """
    if decimal == 0 and digit < 0:
        return f"{0:.{-digit}f}"
    if decimal in [np.inf, np.nan, -np.inf]:
        return str(decimal)
    order = order_of_magnitude(decimal)
    number_of_digits = order - digit
    if number_of_digits < max_string_length and -DEFAULT_MAX_STRING_LENGTH < digit < 0:
        return f"{decimal:.{-digit}f}"
    order_sign = "+" if order > 0 else "-"
    normalized_value = decimal * np.power(10.0, -order)
    return f"{normalized_value:.{number_of_digits}f}e{order_sign}{np.abs(order)}"


def order_of_magnitude(decimal: float) -> int:
    """Get the order of magnitude of the given number."""
    if decimal == 0 or is_nan_or_infinity(decimal):
        return 0
    return int(np.floor(np.log10(np.abs(decimal))))


def is_nan_or_infinity(decimal: float) -> bool:
    """Whether a floating number is infinity or NaN (Not a Number)."""
    return np.isnan(decimal) or np.isinf(decimal)



"""Fit-random CLI method."""
from pathlib import Path
from typing import Optional, Union

import click

from eddington.cli.common_flags import (
    a0_option,
    fitting_function_argument,
    is_grid_option,
    is_json_option,
    is_legend_option,
    is_x_log_scale_option,
    is_y_log_scale_option,
    output_dir_option,
    polynomial_option,
    should_plot_data_option,
    should_plot_fitting_option,
    should_plot_residuls_option,
    title_option,
    x_label_option,
    y_label_option,
)
from eddington.cli.main_cli import eddington_cli
from eddington.cli.util import (
    extract_array_from_string,
    fit_and_plot,
    load_fitting_function,
)
from eddington.consts import (
    DEFAULT_MAX_COEFF,
    DEFAULT_MIN_COEFF,
    DEFAULT_XMAX,
    DEFAULT_XMIN,
    DEFAULT_XSIGMA,
    DEFAULT_YSIGMA,
)
from eddington.fitting_data import FittingData

# pylint: disable=invalid-name,too-many-arguments,too-many-locals,duplicate-code


@eddington_cli.command("fit-random")
@click.pass_context
@fitting_function_argument
@polynomial_option
@a0_option
@click.option(
    "--a",
    type=str,
    help="Actual parameters of the randomized data.",
)
@click.option(
    "--random-x-min",
    type=float,
    default=DEFAULT_XMIN,
    help="Minimum x value for the random data",
)
@click.option(
    "--random-x-max",
    type=float,
    default=DEFAULT_XMAX,
    help="Maximum x value for the random data",
)
@click.option(
    "--x-sigma",
    type=float,
    default=DEFAULT_XSIGMA,
    help="Standard deviation of the random x values",
)
@click.option(
    "--y-sigma",
    type=float,
    default=DEFAULT_YSIGMA,
    help="Standard deviation of the random y values",
)
@click.option(
    "--min-coeff",
    type=float,
    default=DEFAULT_MIN_COEFF,
    help="Minimum value for the parameters",
)
@click.option(
    "--max-coeff",
    type=float,
    default=DEFAULT_MAX_COEFF,
    help="Maximum value for the parameters",
)
@title_option
@x_label_option
@y_label_option
@is_grid_option
@should_plot_fitting_option
@should_plot_residuls_option
@should_plot_data_option
@is_legend_option
@is_x_log_scale_option
@is_y_log_scale_option
@output_dir_option
@is_json_option
def eddington_fit_random(
    ctx: click.Context,
    fitting_function_name: Optional[str],
    polynomial_degree: Optional[int],
    a0: Optional[str],
    a: Optional[str],
    random_x_min: float,
    random_x_max: float,
    min_coeff: float,
    max_coeff: float,
    x_sigma: float,
    y_sigma: float,
    title: Optional[str],
    x_label: Optional[str],
    y_label: Optional[str],
    grid,
    should_plot_fitting: bool,
    should_plot_residuals: bool,
    should_plot_data: bool,
    legend: Optional[bool],
    x_log_scale: bool,
    y_log_scale: bool,
    output_dir: Union[Path, str],
    json: bool,
):
    """
    Fitting random data using the Eddington fitting algorithm.

    This is best used for testing fitting functions.
    """
    func = load_fitting_function(
        ctx=ctx, func_name=fitting_function_name, polynomial_degree=polynomial_degree
    )
    data = FittingData.random(
        func,
        xmin=random_x_min,
        xmax=random_x_max,
        min_coeff=min_coeff,
        max_coeff=max_coeff,
        xsigma=x_sigma,
        ysigma=y_sigma,
        a=extract_array_from_string(a),
    )
    fit_and_plot(
        data=data,
        func=func,
        a0=extract_array_from_string(a0),
        legend=legend,
        output_dir=output_dir,
        is_json=json,
        title=title,
        x_label=x_label,
        y_label=y_label,
        x_log_scale=x_log_scale,
        y_log_scale=y_log_scale,
        grid=grid,
        should_plot_data=should_plot_data,
        should_plot_fitting=should_plot_fitting,
        should_plot_residuals=should_plot_residuals,
    )



#!/usr/bin/env python
import os

from setuptools import setup


version = "0.0.23"
install_requires = [
    "click >= 7.1.2",
    "prettytable >= 2.0.0",
    'dataclasses >= 0.7; python_version == "3.6"',
    "scipy >= 1.5.4",
    "numpy == 1.19.3",
    "openpyxl >= 3.0.5",
    "matplotlib >= 3.3.3",
]

if os.environ.get("READTHEDOCS") == "True":
    install_requires = install_requires[:3]

setup(version=version, install_requires=install_requires)



"""Utility methods for the Eddington CLI."""
import re
from pathlib import Path
from typing import Optional

import click
import numpy as np

from eddington.fitting import fit
from eddington.fitting_data import FittingData
from eddington.fitting_functions_list import linear, polynomial
from eddington.fitting_functions_registry import FittingFunctionsRegistry
from eddington.fitting_result import FittingResult
from eddington.plot import plot_data, plot_fitting, plot_residuals, show_or_export


def load_data_file(ctx: click.Context, data_file: Path, sheet: Optional[str], **kwargs):
    """Load data file with any suffix."""
    suffix = data_file.suffix
    if suffix == ".csv":
        return FittingData.read_from_csv(filepath=data_file, **kwargs)
    if suffix == ".json":
        return FittingData.read_from_json(filepath=data_file, **kwargs)
    if suffix != ".xlsx":
        click.echo(f'Cannot read data with "{suffix}" suffix')
        ctx.exit(1)
    if sheet is None:
        click.echo("Sheet name has not been specified!")
        ctx.exit(1)
    return FittingData.read_from_excel(filepath=data_file, sheet=sheet, **kwargs)


def load_fitting_function(
    ctx: click.Context, func_name: Optional[str], polynomial_degree: Optional[int]
):
    """Load appropriate fitting function."""
    if func_name == "":
        if polynomial_degree is not None:
            return polynomial(polynomial_degree)
        return linear
    if polynomial_degree is not None:
        click.echo("Cannot accept both polynomial and fitting function")
        ctx.exit(1)
    return FittingFunctionsRegistry.load(func_name)


def fit_and_plot(  # pylint: disable=too-many-arguments,invalid-name
    data,
    func,
    a0,
    legend,
    output_dir,
    is_json,
    title,
    x_label,
    y_label,
    should_plot_data,
    should_plot_fitting,
    should_plot_residuals,
    **plot_kwargs,
):
    """Plot everything needed."""
    output_dir = None if output_dir is None else Path(output_dir)
    result = fit(data, func, a0=a0)
    write_and_export_result(
        result, func_name=func.name, output_dir=output_dir, is_json=is_json
    )
    title = func.name.title() if title is None else title
    x_label = data.x_column if x_label is None else x_label
    y_label = data.y_column if y_label is None else y_label
    if should_plot_data:
        with plot_data(
            data=data,
            title_name=f"{title} - Data",
            xlabel=x_label,
            ylabel=y_label,
            **plot_kwargs,
        ) as fig:
            show_or_export(
                fig, output_path=__optional_path(output_dir, f"{func.name}_data.png")
            )
    if should_plot_fitting:
        with plot_fitting(
            func=func,
            data=data,
            a=result.a,
            title_name=f"{title}",
            legend=legend,
            xlabel=x_label,
            ylabel=y_label,
            **plot_kwargs,
        ) as fig:
            show_or_export(
                fig, output_path=__optional_path(output_dir, f"{func.name}.png")
            )
    if should_plot_residuals:
        with plot_residuals(
            func=func,
            data=data,
            a=result.a,
            title_name=f"{title} - Residuals",
            xlabel=x_label,
            ylabel=y_label,
            **plot_kwargs,
        ) as fig:
            show_or_export(
                fig,
                output_path=__optional_path(output_dir, f"{func.name}_residuals.png"),
            )


def extract_array_from_string(a0: Optional[str]):  # pylint: disable=invalid-name
    """Split a0 string to an array."""
    if a0 is None:
        return None
    return np.array(list(map(float, re.split(",[ \t]*", a0))))


def write_and_export_result(
    result: FittingResult,
    func_name: str,
    output_dir: Optional[Path] = None,
    is_json: bool = False,
):
    """Write result to console and to file if specified so."""
    click.echo(result.pretty_string)
    if output_dir is None:
        return
    output_dir.mkdir(parents=True, exist_ok=True)
    if is_json:
        result.save_json(output_dir / f"{func_name}_result.json")
    else:
        result.save_txt(output_dir / f"{func_name}_result.txt")


def __optional_path(directory: Optional[Path], file_name: str):
    if directory is None:
        return None
    return directory / file_name



"""Main CLI method."""

import click

from eddington import __version__


@click.group("eddington")
@click.version_option(version=__version__)
def eddington_cli():
    """Command line for Eddington."""
